class ResolverTargetAttrNotFound(Exception):
    pass

class DataloaderDependCantBeResolved(Exception):
    pass

class LoaderFieldNotProvidedError(Exception):
    pass

class MissingAnnotationError(Exception):
    pass